from typing import Dict, Any

from unsserv.common.utils import IConfig


class ABloomConfig(IConfig):
    DEPTH = 2
    TIMEOUT = 10
    MAINTENANCE_SLEEP = 1

    def load_from_dict(self, config_dict: Dict[str, Any]):
        self.DEPTH = config_dict.get("depth", ABloomConfig.DEPTH)
        self.TIMEOUT = config_dict.get("timeout", ABloomConfig.TIMEOUT)
        self.MAINTENANCE_SLEEP = config_dict.get(
            "maintenance_sleep", ABloomConfig.MAINTENANCE_SLEEP
        )



from typing import List, Tuple, Union
from unsserv.common.structs import Node

EventData = bytes
EventId = str
EventOrigin = Node
LpbcastEvent = List[Union[EventId, EventData, EventOrigin]]
Digest = List[Tuple[EventId, EventOrigin]]



from typing import Dict, Any

from unsserv.common.utils import IConfig


class LpbcastConfig(IConfig):
    FANOUT = 10
    THRESHOLD = 10

    def load_from_dict(self, config_dict: Dict[str, Any]):
        self.FANOUT = config_dict.get("fanout", LpbcastConfig.FANOUT)
        self.THRESHOLD = config_dict.get("threhold", LpbcastConfig.THRESHOLD)



from typing import Callable, Any, List, Dict

from unsserv.common.typing import View
from unsserv.common.structs import Node

CustomSelectionRanking = Callable[[View], List[Node]]
ExternalViewSource = Callable
Payload = Dict[Any, Any]



from collections import Counter
from typing import Any, Callable, Coroutine, Union

View = Counter

SyncHandler = Callable[..., None]
AsyncHandler = Callable[..., Coroutine[Any, Any, None]]
Handler = Union[SyncHandler, AsyncHandler]



from abc import ABC, abstractmethod
from typing import Any, List, Optional, Set

from unsserv.common.structs import Node, Property
from unsserv.common.typing import Handler


class IService(ABC):
    """
    UnsServ service interface.

    :ivar my_node: service node
    :ivar service_id: service identifier
    :ivar running: boolean that represents if the service is running
    """

    @abstractmethod
    async def join(self, service_id: Any, **configuration: Any):
        """
        Join/start service.

        :param service_id: service's id
        :param configuration: configuration parameters.
        :return:
        """
        pass

    @abstractmethod
    async def leave(self):
        """
        Leave/stop service.

        :return:
        """
        pass

    properties: Set[Property]
    my_node: Node
    service_id: Any
    running: bool = False


class ISubService(IService):
    """
    Interface of UnsServ service that is not the first layer (check
    documentation's stack).

    :ivar membership: Membership service underlying (from the layer below) the service.
    """

    membership: "IMembershipService"


class IMembershipService(IService):
    """Membership service interface."""

    @abstractmethod
    def get_neighbours(self) -> List[Node]:
        """
        Get the neighbours to which the service is connected.

        as a list or a View.
        :return: neighbours to which the service is connected.
        """
        pass

    @abstractmethod
    def add_neighbours_handler(self, handler: Handler):
        """
        Add a handler that is executed when neighbours change.

        :param handler: function that will be called.
        as a list or a View.
        :return:
        """
        pass

    @abstractmethod
    def remove_neighbours_handler(self, handler: Handler):
        """
        Remove the handler that is executed when neighbours change.

        :param handler: handler that will be removed.
        :return:
        """
        pass


class IClusteringService(ISubService, IMembershipService):
    """Clustering service for having bias when selecting neighbours."""

    pass


class IAggregationService(ISubService):
    """Aggregation service."""

    @abstractmethod
    async def get_aggregate(self) -> Any:
        """
        Get current aggregate value.

        :return: aggregate value.
        """
        pass

    @abstractmethod
    def add_aggregate_handler(self, handler: Handler):
        pass

    @abstractmethod
    def remove_aggregate_handler(self, handler: Handler):
        pass


class ISamplingService(ISubService):
    """Sampling service."""

    @abstractmethod
    async def get_sample(self) -> Node:
        """
        Gets a sample.

        :return: sample Node.
        """
        pass


class IDisseminationService(ISubService):
    @abstractmethod
    async def broadcast(self, data: bytes):
        """
        Disseminates/brodcasts data.

        :param data: bytes data
        :return:
        """
        pass

    @abstractmethod
    def add_broadcast_handler(self, handler: Handler):
        pass

    @abstractmethod
    def remove_broadcast_handler(self, handler: Handler):
        pass


class ISearchingService(ISubService):
    @abstractmethod
    async def publish(self, data_id: str, data: bytes):
        """
        Publish data that can be found when searching.

        :param data_id: data id.
        :param data: bytes data.
        :return:
        """
        pass

    @abstractmethod
    async def unpublish(self, data_id: str):
        """
        Unpublish data so that cannot be found when searching.

        :param data_id: data id.
        :return:
        """
        pass

    @abstractmethod
    async def search(self, data_id: str) -> Optional[bytes]:
        """
        Search data.

        :param data_id: data id.
        :return: Data bytes or None if not found.
        """
        pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    "rpcudp==4.0.0",
]

extras_require = {
    "dev": [
        "pytest==5.4.1",
        "pytest-asyncio==0.11.0",
        "mypy==0.740",
        "flake8==3.7.9",
        "black==19.10b0"
    ],
}

setuptools.setup(
    name="unsserv",
    version="0.0.1",
    author="Aratz M. Lasa",
    author_email="aratz.m.lasa@gmail.com",
    description="P2P high-level library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aratz-lasa/py-unsserv",
    install_requires=install_requires,
    extras_require=extras_require,
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



import asyncio

import pytest

from tests.utils import init_stable_membership
from unsserv.common.gossip.config import GossipConfig
from unsserv.stable.sampling.rwd import RWD

init_stable_membership = init_stable_membership  # for flake8 compliance

AGGREGATION_SERVICE_ID = "rwd"


@pytest.mark.asyncio
@pytest.fixture
async def init_rwd():
    sampl = None
    r_sampls = []

    async def _init_rwd(hypa, r_hypas):
        nonlocal sampl, r_sampls
        sampl = RWD(hypa)
        await sampl.join(AGGREGATION_SERVICE_ID)
        for r_hypa in r_hypas:
            r_sampl = RWD(r_hypa)
            await r_sampl.join(AGGREGATION_SERVICE_ID)
            r_sampls.append(r_sampl)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return r_sampls, sampl

    try:
        yield _init_rwd
    finally:
        await sampl.leave()
        for r_sampl in r_sampls:
            await r_sampl.leave()


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_start_stop(init_stable_membership, init_rwd, amount):
    hypa, r_hypas = await init_stable_membership(amount)
    r_sampls, sampl = await init_rwd(hypa, r_hypas)


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_sampling(init_stable_membership, init_rwd, amount):
    hypa, r_hypas = await init_stable_membership(amount)
    r_sampls, sampl = await init_rwd(hypa, r_hypas)

    samples = {await sampl.get_sample() for _ in range(amount * 2)}
    assert len({r_hypa.my_node for r_hypa in r_hypas} - samples) / len(r_hypas) <= 0.45



import asyncio

import pytest

from tests.utils import init_extreme_membership
from unsserv.common.aggregation.anti_entropy import (
    AntiEntropy,
    aggregate_functions,
)
from unsserv.common.aggregation.config import AggregateType
from unsserv.common.gossip.config import GossipConfig

init_extreme_membership = init_extreme_membership  # for flake8 compliance

AGGR_SERVICE_ID = "tman"


@pytest.mark.asyncio
@pytest.fixture
async def init_anti_entropy():
    anti = None
    r_antis = []

    async def _init_anti_entropy(newc, r_newcs):
        nonlocal anti, r_antis
        anti = AntiEntropy(newc)
        await anti.join(
            AGGR_SERVICE_ID,
            aggregate_type=AggregateType.MEAN,
            aggregate_value=anti.my_node.address_info[1],
        )
        for r_newc in r_newcs:
            r_anti = AntiEntropy(r_newc)
            await r_anti.join(
                AGGR_SERVICE_ID,
                aggregate_type=AggregateType.MEAN,
                aggregate_value=r_newc.my_node.address_info[1],
            )
            r_antis.append(r_anti)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return anti, r_antis

    try:
        yield _init_anti_entropy
    finally:
        await anti.leave()
        for r_anti in r_antis:
            await r_anti.leave()


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_start_stop(init_extreme_membership, init_anti_entropy, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    anti, r_antis = await init_anti_entropy(newc, r_newcs)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount,aggregate_type",
    [
        (amount, aggregate_type)
        for amount in [1, 5, 100]
        for aggregate_type in AggregateType
    ],
)
async def test_aggregate(
    init_extreme_membership, init_anti_entropy, amount, aggregate_type
):
    newc, r_newcs = await init_extreme_membership(amount)
    anti, r_antis = await init_anti_entropy(newc, r_newcs)

    first_port = anti.my_node.address_info[1]
    assert (
        abs(
            await anti.get_aggregate()
            - aggregate_functions[aggregate_type](
                [number + first_port for number in range(amount + 1)]
            )
        )
        / await anti.get_aggregate()
        < 0.1
    )
    for r_anti in r_antis:
        assert (
            abs(
                await r_anti.get_aggregate()
                - aggregate_functions[aggregate_type](
                    [number + first_port for number in range(amount + 1)]
                )
            )
            / await r_anti.get_aggregate()
            < 0.1
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount",
    [
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 1,
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 5,
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 100,
    ],
)
async def test_aggregate_handler(init_extreme_membership, init_anti_entropy, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    anti, r_antis = await init_anti_entropy(newc, r_newcs)

    handler_event = asyncio.Event()

    async def handler(neighbours):
        nonlocal handler_event
        handler_event.set()

    anti.add_aggregate_handler(handler)

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
    assert handler_event.is_set()



from typing import List, Any, Optional

from unsserv.common.gossip.gossip import Gossip
from unsserv.common.services_abc import IMembershipService
from unsserv.common.structs import Node, Property
from unsserv.common.typing import Handler
from unsserv.stable.membership.double_layered.double_layered import IDoubleLayered
from unsserv.stable.membership.double_layered.config import DoubleLayeredConfig


class HyParView(IMembershipService, IDoubleLayered):
    properties = {Property.STABLE, Property.SYMMETRIC, Property.HAS_GOSSIP}
    gossip: Optional[Gossip]

    def __init__(self, my_node: Node):
        super().__init__(my_node)
        self.gossip = None
        self._config = DoubleLayeredConfig()

    async def join(self, service_id: Any, **configuration: Any):
        if self.running:
            raise RuntimeError("Already running Membership")
        self.service_id = service_id
        self.gossip = Gossip(
            my_node=self.my_node,
            service_id=f"gossip-{service_id}",
            local_view_nodes=configuration.get("bootstrap_nodes", None),
        )
        await self.gossip.start()
        await self._start_two_layered(f"double_layered-{service_id}")
        self.running = True

    async def leave(self):
        if not self.running:
            return
        await self.gossip.stop()
        await self._stop_two_layered()
        self.running = False

    def get_neighbours(self) -> List[Node]:
        return list(self._active_view)

    def add_neighbours_handler(self, handler: Handler):
        if not self.running:
            raise RuntimeError("Membership service not running")
        self._handler_manager.add_handler(handler)

    def remove_neighbours_handler(self, handler: Handler):
        self._handler_manager.remove_handler(handler)

    def _get_passive_view_nodes(self):
        return list(self.gossip.local_view.keys())



from enum import IntEnum, auto
from typing import Tuple, Sequence, Dict, Any

from unsserv.common.rpc.protocol import AProtocol, ITranscoder, Command, Data, Handler
from unsserv.common.structs import Node
from unsserv.common.rpc.structs import Message
from unsserv.common.utils import parse_node
from unsserv.stable.membership.double_layered.structs import ForwardJoin

MessageData = Dict[str, Any]

FIELD_COMMAND = "double_layered-command"
FIELD_TTL = "double_layered-ttl"
FIELD_PRIORITY = "double_layered-priority"
FIELD_ORIGIN_NODE = "double_layered-origin-node"


class DoubleLayeredCommand(IntEnum):
    JOIN = auto()
    FORWARD_JOIN = auto()
    CONNECT = auto()  # it is equivalent to NEIGHBOR in DoubleLayered specification
    DISCONNECT = auto()
    STAY_CONNECTED = auto()


class DoubleLayeredTranscoder(ITranscoder):
    def encode(self, command: Command, *data: Data) -> Message:
        message_data: MessageData  # needed for mypy compliance
        if command == DoubleLayeredCommand.JOIN:
            message_data = {FIELD_COMMAND: DoubleLayeredCommand.JOIN}
            return Message(self.my_node, self.service_id, message_data)
        elif command == DoubleLayeredCommand.FORWARD_JOIN:
            forward_join: ForwardJoin = data[0]
            message_data = {
                FIELD_COMMAND: DoubleLayeredCommand.FORWARD_JOIN,
                FIELD_ORIGIN_NODE: forward_join.origin_node,
                FIELD_TTL: forward_join.ttl,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == DoubleLayeredCommand.CONNECT:
            is_a_priority: bool = data[0]
            message_data = {
                FIELD_COMMAND: DoubleLayeredCommand.CONNECT,
                FIELD_PRIORITY: is_a_priority,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == DoubleLayeredCommand.DISCONNECT:
            message_data = {FIELD_COMMAND: DoubleLayeredCommand.DISCONNECT}
            return Message(self.my_node, self.service_id, message_data)
        elif command == DoubleLayeredCommand.STAY_CONNECTED:
            message_data = {FIELD_COMMAND: DoubleLayeredCommand.STAY_CONNECTED}
            return Message(self.my_node, self.service_id, message_data)
        raise ValueError("Invalid Command")

    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        command = message.data[FIELD_COMMAND]
        if command == DoubleLayeredCommand.JOIN:
            return DoubleLayeredCommand.JOIN, []
        elif command == DoubleLayeredCommand.FORWARD_JOIN:
            forward_join = ForwardJoin(
                origin_node=parse_node(message.data[FIELD_ORIGIN_NODE]),
                ttl=message.data[FIELD_TTL],
            )
            return DoubleLayeredCommand.FORWARD_JOIN, [forward_join]
        elif command == DoubleLayeredCommand.CONNECT:
            is_a_priority: bool = message.data[FIELD_PRIORITY]
            return DoubleLayeredCommand.CONNECT, [is_a_priority]
        elif command == DoubleLayeredCommand.DISCONNECT:
            return DoubleLayeredCommand.DISCONNECT, []
        elif command == DoubleLayeredCommand.STAY_CONNECTED:
            return DoubleLayeredCommand.STAY_CONNECTED, []
        raise ValueError("Invalid Command")


class DoubleLayeredProtocol(AProtocol):
    def _get_new_transcoder(self):
        return DoubleLayeredTranscoder(self.my_node, self.service_id)

    async def join(self, destination: Node):
        message = self._transcoder.encode(DoubleLayeredCommand.JOIN)
        return await self._rpc.call_send_message(destination, message)

    async def forward_join(self, destination: Node, forward_join: ForwardJoin):
        message = self._transcoder.encode(
            DoubleLayeredCommand.FORWARD_JOIN, forward_join
        )
        return await self._rpc.call_send_message(destination, message)

    async def connect(self, destination: Node, is_a_priority: bool) -> int:
        message = self._transcoder.encode(DoubleLayeredCommand.CONNECT, is_a_priority)
        return await self._rpc.call_send_message(destination, message)

    async def disconnect(self, destination: Node) -> int:
        message = self._transcoder.encode(DoubleLayeredCommand.DISCONNECT)
        return await self._rpc.call_send_message(destination, message)

    async def stay_connected(self, destination: Node) -> int:
        message = self._transcoder.encode(DoubleLayeredCommand.STAY_CONNECTED)
        return await self._rpc.call_send_message(destination, message)

    def set_handler_join(self, handler: Handler):
        self._handlers[DoubleLayeredCommand.JOIN] = handler

    def set_handler_forward_join(self, handler: Handler):
        self._handlers[DoubleLayeredCommand.FORWARD_JOIN] = handler

    def set_handler_connect(self, handler: Handler):
        self._handlers[DoubleLayeredCommand.CONNECT] = handler

    def set_handler_disconnect(self, handler: Handler):
        self._handlers[DoubleLayeredCommand.DISCONNECT] = handler

    def set_handler_stay_connected(self, handler: Handler):
        self._handlers[DoubleLayeredCommand.STAY_CONNECTED] = handler



import asyncio
import math
import random
from typing import Any, Set

from unsserv.common.errors import ServiceError
from unsserv.common.services_abc import IDisseminationService, IMembershipService
from unsserv.common.structs import Node, Property
from unsserv.common.typing import Handler
from unsserv.common.utils import get_random_id, HandlersManager, stop_task
from unsserv.stable.dissemination.brisa.config import BrisaConfig
from unsserv.stable.dissemination.brisa.protocol import BrisaProtocol
from unsserv.stable.dissemination.brisa.typing import BroadcastLevel


class Brisa(IDisseminationService):
    properties = {Property.STABLE, Property.ONE_TO_MANY}
    _protocol: BrisaProtocol
    _handler_manager: HandlersManager
    _config: BrisaConfig

    _level: BroadcastLevel
    _children: Set[Node]
    _parents: Set[Node]
    _im_root: bool
    _maintenance_task: asyncio.Task

    def __init__(self, membership: IMembershipService):
        self.my_node = membership.my_node
        self.membership = membership
        self._protocol = BrisaProtocol(self.my_node)
        self._handler_manager = HandlersManager()
        self._config = BrisaConfig()

        self._children = set()
        self._parents = set()
        self._level = math.inf
        self._im_root = False

    async def join(self, service_id: str, **configuration: Any):
        if self.running:
            raise RuntimeError("Already running Dissemination")
        self.service_id = service_id
        self._im_root = configuration.get("im_root", False)
        await self._initialize_protocol()
        self._handler_manager.add_handler(configuration["broadcast_handler"])
        self._config.load_from_dict(configuration)
        self._maintenance_task = asyncio.create_task(self._maintain_dag_loop())
        self.running = True

    async def leave(self):
        if not self.running:
            return
        await stop_task(self._maintenance_task)
        self._handler_manager.remove_all_handlers()
        await self._protocol.stop()
        self.running = False

    async def broadcast(self, data: bytes):
        if not self.running:
            raise RuntimeError("Dissemination service not running")
        if self._im_root is False:
            raise RuntimeError("Node must be root to broadcast")
        assert isinstance(data, bytes)
        await asyncio.wait_for(self._disseminate(data), timeout=self._config.TIMEOUT)

    def add_broadcast_handler(self, handler: Handler):
        self._handler_manager.add_handler(handler)

    def remove_broadcast_handler(self, handler: Handler):
        self._handler_manager.remove_handler(handler)

    async def _maintain_dag_loop(self):
        if self._im_root:
            self._broadcast_id = get_random_id()
            self._level = 0
        while True:
            await asyncio.sleep(self._config.MAINTENANCE_SLEEP)
            if not self._im_root:
                await self._maintain_parents()
                if not self._parents:
                    continue
            await self._maintain_children()

    async def _maintain_parents(self):
        for parent in self._parents.copy():
            if not await self._protocol.im_your_child(parent):
                self._parents.remove(parent)
        if not self._parents:
            await self._find_parent()

    async def _find_parent(self):
        self._level = math.inf
        candidate_parents = self.membership.get_neighbours()
        while not self._parents and candidate_parents:
            parent = random.choice(candidate_parents)
            candidate_parents.remove(parent)
            try:
                level = await self._protocol.become_my_parent(parent)
                if level == -1:
                    continue
                self._level = min(self._level, level + 1)
                self._parents.add(parent)
            except ConnectionError:
                pass

    async def _maintain_children(self):
        for child in self._children.copy():
            try:
                if not await self._protocol.session(child, self._level):
                    self._children.remove(child)
            except ConnectionError:
                pass
        if len(self._children) < self._config.FANOUT:
            await self._find_children()

    async def _find_children(self):
        candidate_children = self.membership.get_neighbours()
        while len(self._children) < self._config.FANOUT and candidate_children:
            child = random.choice(candidate_children)
            candidate_children.remove(child)
            try:
                if await self._protocol.session(child, self._level):
                    self._children.add(child)
            except ConnectionError:
                pass

    async def _disseminate(self, data: bytes):
        pushed_amount = 0
        for child in self._children:
            try:
                await self._protocol.push(child, data)
                pushed_amount += 1
            except ConnectionError:
                pass
        if pushed_amount == 0:
            raise ServiceError("Unable to peer with neighbours for disseminating")

    async def _handler_session(self, sender: Node, level: BroadcastLevel):
        if self._level <= level:
            return False
        self._parents.add(sender)
        self._level = min(level + 1, self._level)
        return True

    async def _handler_push(self, sender: Node, data: bytes):
        asyncio.create_task(self._disseminate(data))
        self._handler_manager.call_handlers(data)

    async def _handler_im_your_child(self, sender: Node):
        return sender in self._children

    async def _handler_become_my_parent(self, sender: Node):
        if (self._im_root or self._parents) and len(
            self._children
        ) < self._config.FANOUT:
            self._children.add(sender)
            return self._level
        return -1

    async def _initialize_protocol(self):
        self._protocol.set_handler_session(self._handler_session)
        self._protocol.set_handler_push(self._handler_push)
        self._protocol.set_handler_im_your_child(self._handler_im_your_child)
        self._protocol.set_handler_become_my_parent(self._handler_become_my_parent)
        await self._protocol.start(self.service_id)
